[v4,2/2] eal: initialize shared plugins on Windows

Message ID 1704757091-7964-3-git-send-email-roretzla@linux.microsoft.com (mailing list archive)
State Superseded, archived
Delegated to: Thomas Monjalon
Headers
Series eal: initialize shared plugins on Windows |

Checks

Context Check Description
ci/checkpatch warning coding style issues
ci/Intel-compilation warning apply issues

Commit Message

Tyler Retzlaff Jan. 8, 2024, 11:38 p.m. UTC
  When EAL is built with MSVC it is possible to dynamically load plugins
on Windows. Hook eal_plugins_init into rte_eal_init if built with MSVC
and provide code to load plugins on Windows.

Signed-off-by: Tyler Retzlaff <roretzla@linux.microsoft.com>
---
 lib/eal/common/eal_common_options.c | 90 ++++++++++++++++++++++++++++++-------
 lib/eal/windows/eal.c               |  8 ++++
 2 files changed, 83 insertions(+), 15 deletions(-)
  

Comments

Dmitry Kozlyuk Jan. 25, 2024, 11:04 p.m. UTC | #1
2024-01-08 15:38 (UTC-0800), Tyler Retzlaff:
> When EAL is built with MSVC it is possible to dynamically load plugins
> on Windows. Hook eal_plugins_init into rte_eal_init if built with MSVC
> and provide code to load plugins on Windows.

Windows implementation does not check directory permissions like its Unix
counterpart does. I don't insist on porting it, because this feature is not
documented and is more like a safety net.
Just making sure this is a conscious decision and not an oversight.

Acked-by: Dmitry Kozlyuk <dmitry.kozliuk@gmail.com>
  

Patch

diff --git a/lib/eal/windows/eal.c b/lib/eal/windows/eal.c
index 52f0e74..dc2bcb7 100644
--- a/lib/eal/windows/eal.c
+++ b/lib/eal/windows/eal.c
@@ -305,6 +305,14 @@  enum rte_proc_type_t
 	if (fctret < 0)
 		exit(1);
 
+#ifdef RTE_TOOLCHAIN_MSVC
+	if (eal_plugins_init() < 0) {
+		rte_eal_init_alert("Cannot init plugins");
+		rte_errno = EINVAL;
+		return -1;
+	}
+#endif
+
 	if (eal_option_device_parse()) {
 		rte_errno = ENODEV;
 		return -1;
diff --git a/lib/eal/common/eal_common_options.c b/lib/eal/common/eal_common_options.c
index d974807..d519a55 100644
--- a/lib/eal/common/eal_common_options.c
+++ b/lib/eal/common/eal_common_options.c
@@ -18,9 +18,7 @@ 
 #include <libgen.h>
 #endif
 #include <sys/stat.h>
-#ifndef RTE_EXEC_ENV_WINDOWS
 #include <dirent.h>
-#endif
 
 #include <rte_string_fns.h>
 #include <rte_eal.h>
@@ -123,10 +121,8 @@  struct shared_driver {
 static struct shared_driver_list solib_list =
 TAILQ_HEAD_INITIALIZER(solib_list);
 
-#ifndef RTE_EXEC_ENV_WINDOWS
 /* Default path of external loadable drivers */
 static const char *default_solib_dir = RTE_EAL_PMD_PATH;
-#endif
 
 /*
  * Stringified version of solib path used by dpdk-pmdinfo.py
@@ -371,12 +367,12 @@  struct device_option {
 }
 
 #ifdef RTE_EXEC_ENV_WINDOWS
-int
-eal_plugins_init(void)
-{
-	return 0;
-}
+#define SOEXT    ".dll"
 #else
+#define SOEXT    ".so"
+#endif
+
+#define SOABIEXT  SOEXT"."ABI_VERSION
 
 static int
 eal_plugindir_init(const char *path)
@@ -397,12 +393,14 @@  struct device_option {
 
 	while ((dent = readdir(d)) != NULL) {
 		struct stat sb;
-		int nlen = strnlen(dent->d_name, sizeof(dent->d_name));
+		size_t nlen = strnlen(dent->d_name, sizeof(dent->d_name));
 
-		/* check if name ends in .so or .so.ABI_VERSION */
-		if (strcmp(&dent->d_name[nlen - 3], ".so") != 0 &&
-		    strcmp(&dent->d_name[nlen - 4 - strlen(ABI_VERSION)],
-			   ".so."ABI_VERSION) != 0)
+		if (nlen < strlen(SOABIEXT))
+			continue;
+
+		/* check if name ends in SOEXT or SOABIEXT */
+		if (strcmp(&dent->d_name[nlen - strlen(SOEXT)], SOEXT) != 0 &&
+		    strcmp(&dent->d_name[nlen - strlen(SOABIEXT)], SOABIEXT) != 0)
 			continue;
 
 		snprintf(sopath, sizeof(sopath), "%s/%s", path, dent->d_name);
@@ -420,6 +418,68 @@  struct device_option {
 	return (dent == NULL) ? 0 : -1;
 }
 
+#ifdef RTE_EXEC_ENV_WINDOWS
+static void*
+eal_dlopen(const char *pathname)
+{
+	void *retval = NULL;
+	struct stat pathstat;
+	char *fullpath = _fullpath(NULL, pathname, 0);
+
+	const char *loadpath = fullpath;
+	DWORD loadflags = 0;
+
+	if (fullpath == NULL) {
+		RTE_LOG(ERR, EAL, "Error expanding full path for %s, %s\n",
+		pathname, strerror(errno));
+	} else {
+		/* Verify that the path exists */
+		if ((stat(fullpath, &pathstat) != 0) && (errno == ENOENT)) {
+			/* not a full or relative path, try a load from default dirs */
+			loadpath = pathname;
+			loadflags = LOAD_LIBRARY_SEARCH_DEFAULT_DIRS;
+		}
+
+		retval = LoadLibraryExA(loadpath, NULL, loadflags);
+		if (retval == NULL)
+			RTE_LOG(ERR, EAL, "Error loading %s, error code: %lu\n",
+			    loadpath, GetLastError());
+	}
+
+	free(fullpath);
+	return retval;
+}
+
+static int
+is_shared_build(void)
+{
+	int shared = 0;
+	HMODULE apphandle = NULL;
+	HMODULE libhandle = NULL;
+
+	/* if fail to get handle, assume statically linked */
+	apphandle = GetModuleHandleA(NULL);
+	if (apphandle == NULL) {
+		RTE_LOG(ERR, EAL, "Cannot get handle to the app\n");
+		goto out;
+	}
+
+	/* if fail to get handle, assume statically linked */
+	if (GetModuleHandleExA(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS |
+	    GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
+	    (LPCSTR)&eal_plugins_init,
+	    &libhandle)) {
+		if (apphandle != libhandle) {
+			/* lib and app handles are different. */
+			/* Therefore lib is dynamically linked */
+			shared = 1;
+		}
+	}
+
+out:
+	return shared;
+}
+#else
 static int
 verify_perms(const char *dirpath)
 {
@@ -527,6 +587,7 @@  struct device_option {
 	EAL_LOG(INFO, "Detected static linkage of DPDK");
 	return 0;
 }
+#endif
 
 int
 eal_plugins_init(void)
@@ -565,7 +626,6 @@  struct device_option {
 	}
 	return 0;
 }
-#endif
 
 /*
  * Parse the coremask given as argument (hexadecimal string) and fill